/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <gio/gio.h>
#include <sqlite3.h>
#include <sys/time.h>
#include <sys/types.h>

#include <app_manager.h>
#include <runtime_info.h>

#include "bm_plugin_interface.h"
#include "bm_cpu_util.h"
#include "bm_log.h"

#define JIFFY_TO_MS_FACTOR 10	// 1000 / 100
#define JIFFY_TO_MS(x) ((x) >= UINT_MAX / JIFFY_TO_MS_FACTOR ? UINT_MAX : (x) * JIFFY_TO_MS_FACTOR)

struct app_status {
	pid_t pid;
	time_t used_time;	// utime + stime
};

struct cpu_time {
	uint utime;
	uint stime;
};

static time_t last_requested_time;
static struct cpu_time last_cpu_time;

static GDBusConnection *dbus_connection;
static GHashTable *running_app_list;	// Key(appid), Val(app_status)

gint find_app_time(gconstpointer a, gconstpointer b)
{
	const char *appid = (const char *)b;
	const app_time_map_st1 *app_time = (const app_time_map_st1 *)a;

	return g_strcmp0(appid, app_time->app_id);
}

void free_atm_st1(gpointer data)
{
	app_time_map_st1 *atm = (app_time_map_st1 *)data;

	if (!atm)
		return;

	if (atm->app_id)
		free(atm->app_id);
	free(atm);
}

bool is_running_app(const char *appid)
{
	bool is_running = false;
	int ret = app_manager_is_running(appid, &is_running);
	if (ret != APP_MANAGER_ERROR_NONE) {
		_E("app_manager_is_running failed (%d)", ret);
		return false;
	}

	return is_running;
}

int get_cpu_time(struct cpu_time *cpu_time)
{
	int ret;
	FILE *fp = NULL;

	fp = fopen("/proc/stat", "r");
	if (!fp) {
		_E("fopen failed : %m");
		return errno;
	}

	ret = fscanf(fp, "%*s %u %*s %u", &(cpu_time->utime), &(cpu_time->stime));
	fclose(fp);

	if (ret < 2) {
		_E("fscanf failed (%d)", ret);
		return EIO;
	}

	return 0;
}

int deinit()
{
	ENTER;

	if (dbus_connection) {
		g_object_unref(dbus_connection);
		dbus_connection = NULL;
	}

	if (running_app_list) {
		g_hash_table_destroy(running_app_list);
		running_app_list = NULL;
	}

	EXIT;
	return BM_PLUGIN_ERROR_NONE;
}

int init()
{
	ENTER;

	int ret;
	GError *g_err = NULL;

	last_requested_time = time(NULL);
	if (last_requested_time == -1) {
		_E("time failed : %m");
		goto failed;
	}

	ret = get_cpu_time(&last_cpu_time);
	if (ret != 0) {
		_E("Failed to get CPU time (%d)", ret);
		goto failed;
	}

	dbus_connection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &g_err);
	if (!dbus_connection) {
		_E("g_bus_get_sync failed : %s", g_err->message ? g_err->message : NULL);
		g_error_free(g_err);
		goto failed;
	}

	running_app_list = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	if (!running_app_list) {
		_E("g_hash_table_new_full failed");
		goto failed;
	}

	EXIT;
	return BM_PLUGIN_ERROR_NONE;

failed:
	deinit();

	EXIT;
	return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
}

int get_feature_data(bm_data_h *handle, bm_plugin_data_type_e type)
{
	ENTER;

	int ret;

	bm_cpu_st *usage_head = NULL;
	bm_cpu_st *usage = NULL;

	time_t current_time;
	struct cpu_time cur_cpu_time;

	GError *g_err = NULL;
	GDBusMessage *msg_req, *msg_reply;

	/* Sqlite */
	sqlite3 *db;
	sqlite3_stmt *stmt;
	char query[128];

	/* DB row */
	const char *appid;
	const char *data;
	unsigned int utime, stime, pid, status;

	/* App status */
	struct app_status *app_status, *new_app_status;
	int elapsed;
	process_cpu_usage_s *cpu_usage;
	GHashTableIter iter;
	gpointer g_key, g_val;

	/* App time map */
	app_time_map_st1 *app_time_map;


	/* Check argument */
	if (!handle || type != BM_DATA_TYPE_CPU) {
		_E("Invalid argument");
		return BM_PLUGIN_ERROR_INVALID_PARAMETER;
	}

	/* Get current time */
	current_time = time(NULL);
	if (current_time == -1) {
		_E("time failed : %m");

		EXIT;
		return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
	}

	/* Get total CPU time */
	ret = get_cpu_time(&cur_cpu_time);
	if (ret != 0) {
		_E("Failed to get CPU time (%d)", ret);

		EXIT;
		return BM_PLUGIN_ERROR_NO_DATA;
	}

	/* Check total CPU usage */
	if (cur_cpu_time.utime < last_cpu_time.utime || cur_cpu_time.stime < last_cpu_time.stime) {
		_E("Negative CPU usage. Nothing to return");

		EXIT;
		return BM_PLUGIN_ERROR_NO_DATA;
	}

	if (cur_cpu_time.utime == last_cpu_time.utime && cur_cpu_time.stime == last_cpu_time.stime) {
		_I("System doesn't use CPU. Nothing to return");

		EXIT;
		return BM_PLUGIN_ERROR_NO_DATA;
	}

	/* Request for resourced to update HEART-CPU DB */
	msg_req = g_dbus_message_new_method_call(
			"org.tizen.resourced",
			"/Org/Tizen/ResourceD/Logging",
			"org.tizen.resourced.logging",
			"SyncCpuData"
			);
	if (!msg_req) {
		if (g_err) {
			_E("g_dbus_message_new_method_call failed : %s", g_err->message ? g_err->message : NULL);
			g_error_free(g_err);
		} else
			_E("g_dbus_message_new_method_call failed");

		EXIT;
		return BM_PLUGIN_ERROR_NOT_SUPPORTED;
	}

	msg_reply = g_dbus_connection_send_message_with_reply_sync(dbus_connection, msg_req, G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &g_err);
	if (msg_reply)
		g_object_unref(msg_reply);
	else {
		if (g_err) {
			_W("g_dbus_connection_send_message_with_reply_sync failed : %s", g_err->message ? g_err->message : NULL);
			g_error_free(g_err);
		} else
			_W("g_dbus_connection_send_message_with_reply_sync failed");
	}

	/* Open HEART-CPU DB */
	ret = sqlite3_open("/opt/usr/home/owner/.applications/dbspace/.resourced-heart-default.db", &db);
	if (ret != SQLITE_OK) {
		_E("sqlite3_open failed (%d)", ret);

		EXIT;
		return BM_PLUGIN_ERROR_NO_DATA;
	}

	/* Make usage_head data structure */
	usage_head = (bm_cpu_st *)calloc(1, sizeof(bm_cpu_st));
	if (!usage_head) {
		_E("calloc failed");

		EXIT;
		return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
	}

	/* Make usage data structure */
	usage = (bm_cpu_st *)calloc(1, sizeof(bm_cpu_st));
	if (!usage) {
		_E("calloc failed");

		EXIT;
		return BM_PLUGIN_ERROR_OUT_OF_MEMORY;
	}

	/* Update total CPU time */
	usage->utime = JIFFY_TO_MS(cur_cpu_time.utime - last_cpu_time.utime);
	usage->stime = JIFFY_TO_MS(cur_cpu_time.stime - last_cpu_time.stime);

	_D("Total CPU usage : %u", JIFFY_TO_MS(usage->utime + usage->stime));

	/* Read data from HEART-CPU DB */
	_I("Gather CPU usage : %ld ~ %ld", last_requested_time, current_time);
	snprintf(query, 128, "SELECT appid,data FROM cpu WHERE (time >= %ld AND time < %ld)", last_requested_time, current_time);
	ret = sqlite3_prepare_v2(db, query, -1, &stmt, 0);
	while (sqlite3_step(stmt) == SQLITE_ROW) {
		appid = (const char *)sqlite3_column_text(stmt, 0);
		data = (const char *)sqlite3_column_text(stmt, 1);
		sscanf(data, "%u %u %u %u", &utime, &stime, &pid, &status);

		//_D("appid(%s), utime(%u), stime(%u), pid(%u), status(%u)", appid, utime, stime, pid, status);

		/* Ignore terminated app */
		if (!is_running_app(appid)) {
			_D("%s is already terminated", appid);
			g_hash_table_remove(running_app_list, appid);
			continue;
		}

		app_status = g_hash_table_lookup(running_app_list, appid);

		/* Insert newly launched app in the running app list */
		if (!app_status) {
			_D("%s is newly launched", appid);
			new_app_status = malloc(sizeof(struct app_status));
			if (!new_app_status) {
				_E("malloc failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			new_app_status->pid = pid;
			new_app_status->used_time = JIFFY_TO_MS(utime + stime);
			g_hash_table_insert(running_app_list, g_strdup(appid), new_app_status);
			continue;
		}

		/* Initialize restarted app */
		if (app_status->pid != pid) {
			_D("%s is restarted", appid);
			app_status->pid = pid;
			app_status->used_time = JIFFY_TO_MS(utime + stime);
			continue;
		}

		/* Ignore CPU unused app */
		elapsed = JIFFY_TO_MS(utime + stime) - app_status->used_time;
		if (elapsed <= 0) {
			_D("%s doesn't use CPU. Ignore it", appid);
			continue;
		}

		_D("%s uses CPU (+%d ms)", appid, elapsed);

		/* Insert the app using CPU in the ATM list */
		if (!g_slist_find_custom(usage->atm_list, appid, find_app_time)) {
			app_time_map = malloc(sizeof(app_time_map_st1));
			if (!app_time_map) {
				_E("malloc failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->app_id = g_strdup(appid);
			if (!app_time_map->app_id) {
				_E("g_strdup failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->time = elapsed;
			usage->atm_list = g_slist_append(usage->atm_list, app_time_map);
		}

		/* Update app status */
		app_status->used_time += elapsed;
	}

	/* Update all running apps with current value */
	g_hash_table_iter_init(&iter, running_app_list);
	while (g_hash_table_iter_next(&iter, &g_key, &g_val)) {
		appid = (const char *)g_key;
		app_status = (struct app_status *)g_val;

		/* Ignore terminated app */
		if (!is_running_app(appid)) {
			_D("%s is already terminated", appid);
			g_hash_table_iter_remove(&iter);
			continue;
		}

		/* Get latest CPU time */
		ret = runtime_info_get_process_cpu_usage(&app_status->pid, 1, &cpu_usage);
		if (ret != RUNTIME_INFO_ERROR_NONE) {
			_E("runtime_info_get_process_cpu_usage for %u failed(%d)", app_status->pid, ret);
			continue;
		}

		/* Ignore CPU unused app */
		elapsed = JIFFY_TO_MS(cpu_usage->utime + cpu_usage->stime) - app_status->used_time;
		free(cpu_usage);
		if (elapsed <= 0) {
			_D("%s doesn't use CPU. Ignore it", appid);
			continue;
		}

		_D("%s uses CPU (+%d ms)", appid, elapsed);

		/* Insert the app using CPU in the ATM list */
		if (!g_slist_find_custom(usage->atm_list, appid, find_app_time)) {
			app_time_map = malloc(sizeof(app_time_map_st1));
			if (!app_time_map) {
				_E("malloc failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->app_id = g_strdup(appid);
			if (!app_time_map->app_id) {
				_E("g_strdup failed");
				ret = BM_PLUGIN_ERROR_OUT_OF_MEMORY;
				goto clean_atm;
			}
			app_time_map->time = elapsed;
			usage->atm_list = g_slist_append(usage->atm_list, app_time_map);
		}

		app_status->used_time += elapsed;
	}

	/* Update last requested time */
	last_requested_time = current_time;

	/* Update last CPU time */
	last_cpu_time.utime = cur_cpu_time.utime;
	last_cpu_time.stime = cur_cpu_time.stime;

	/* Return result */
	usage_head->cpu_data_list = g_slist_append(usage_head->cpu_data_list, usage);
	*handle = usage_head;

	_I("Succeed to get CPU usage");
	ret = BM_PLUGIN_ERROR_NONE;
	goto clean_db;

clean_atm:
	g_slist_free_full(usage->atm_list, free_atm_st1);

clean_db:
	sqlite3_finalize(stmt);
	sqlite3_close(db);

	EXIT;
	return ret;
}

static bm_api_st cpu_api = {
	.init = init,
	.deinit = deinit,
	.get_feature_data = get_feature_data
};

int plugin_init(bm_api_st **api)
{
	ENTER;

	*api = &cpu_api;

	EXIT;
	return 0;
}

int plugin_deinit(void)
{
	ENTER;

	EXIT;
	return 0;
}

BM_PLUGIN_ADD(BM_PLUGIN_ID_CPU, NAME, AUTHOR, VERSION, plugin_init, plugin_deinit);
